/*************************************************************************
 *
 *  C-to-C Translator
 *
 *  Adapted from Clean ANSI C Parser
 *  Eric A. Brewer, Michael D. Noakes
 * 
 *************************************************************************/
/*
 *
 * Copyright (c) 1994 MIT Laboratory for Computer Science
 * Copyright (c) 2002 Massachusetts Institute of Technology
 * Copyright (c) 2002 Bradley C. Kuszmaul
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *************************************************************************/
#include "ast.h"

FILE_IDENTITY(ident,
	      "$HeadURL: https://bradley.lcs.mit.edu/svn/repos/cilk/5.4.2.2/cilk2c/print-ast.c $ $LastChangedBy: bradley $ $Rev: 1708 $ $Date: 2004-08-02 07:27:43 -0400 (Mon, 02 Aug 2004) $");

#include <ctype.h>

/* main debugging entry points -- handy for calling from gdb */

GLOBAL void DPN(Node *n)
{
     PrintNode(stdout, n, 0);
     putchar('\n');
     fflush(stdout);
}

GLOBAL void DPL(List *list)
{
     PrintList(stdout, list, 0);
     putchar('\n');
     fflush(stdout);
}

GLOBAL void PrintCoord (FILE *out, Coord *c) {
#if 0
    static int prevfilenum=-1;
    fprintf(out, "<");
    if (prevfilenum!=c->file) {
	fprintf(out, "%s:", FileNames[c->file]);
	prevfilenum=c->file;
    }
#endif
    if (PrintASTcoords)
	fprintf(out, " <%d.%d>", c->line, c->offset);
}


/*************************************************************************/
/*                                                                       */
/*                          Expression nodes                             */
/*                                                                       */
/*************************************************************************/

PRIVATE inline void PrintConst(FILE *out, Node *node, UNUSED(ConstNode *u), UNUSED(int offset), UNUSED(Bool norecurse))
{
     fprintf(out, "Const: ");
     PrintConstant(out, node, TRUE);
}

PRIVATE inline void PrintId(FILE *out, Node *node, idNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Id: %s", u->text);
     PrintCoord(out, &node->coord);
     if (u->value) {
	  PrintCRSpaces(out, offset + 2);
	  fputs("Value: ", out);
	  PrintCRSpaces(out, offset + 4);
	  PrintNode(out, u->value, offset + 4);
     }
     PrintCRSpaces(out, offset + 2);
     fputs("decl: ", out);
     PrintNode(out, u->decl, offset+4);
     /* PrintCRSpaces(out, offset + 2);*/
     /* PrintNode(out, u->decl, offset + 2); KHR */
}

PRIVATE inline void PrintBinop(FILE *out, Node *node, binopNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Binop: ");
     PrintOp(out, u->op);
     PrintCoord(out, &node->coord);
     if (u->type) {
	  PrintCRSpaces(out, offset + 2);
	  PrintNode(out, u->type, offset + 2);
     }
     if (u->value) {
	  PrintCRSpaces(out, offset + 2);
	  fputs("Value: ", out);
	  PrintCRSpaces(out, offset + 4);
	  PrintNode(out, u->value, offset + 4);
     }
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->left, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->right, offset + 2);
}

PRIVATE inline void PrintUnary(FILE *out, Node *node, unaryNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Unary: ");
     PrintOp(out, u->op);
     PrintCoord(out, &node->coord);
     if (u->type) {
	  PrintCRSpaces(out, offset + 2);
	  PrintNode(out, u->type, offset + 2);
     }
     if (u->value) {
	  PrintCRSpaces(out, offset + 2);
	  fputs("Value: ", out);
	  PrintCRSpaces(out, offset + 4);
	  PrintNode(out, u->value, offset + 4);
     }
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->expr, offset + 2);
}

PRIVATE inline void PrintCast(FILE *out, UNUSED(Node *node), castNode *u, int offset, UNUSED(Bool norecurse))
{
     fputs("Cast: ", out);
     PrintCoord(out, &node->coord);
     if (u->value) {
	  PrintCRSpaces(out, offset + 2);
	  fputs("Value: ", out);
	  PrintCRSpaces(out, offset + 4);
	  PrintNode(out, u->value, offset + 4);
     }
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->type, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->expr, offset + 2);
}

PRIVATE inline void PrintComma(FILE *out, Node *node, commaNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Comma: List: exprs <%d.%d>", node->coord.line, node->coord.offset);
     if (node->parenthesized) fprintf(out, " parenthesized");
     PrintCRSpaces(out, offset + 2);
     PrintList(out, u->exprs, offset + 2);
}

PRIVATE inline void PrintConstructor(FILE *out, UNUSED(Node *node), constructorNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Constructor: List: exprs");
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->type, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->initializerlist, offset + 2);
}

PRIVATE inline void PrintTernary(FILE *out, UNUSED(Node *node), ternaryNode *u, int offset, UNUSED(Bool norecurse))
{
     fputs("Ternary: ", out);
     if (u->value) {
	  PrintCRSpaces(out, offset + 2);
	  fputs("Value: ", out);
	  PrintCRSpaces(out, offset + 4);
	  PrintNode(out, u->value, offset + 4);
     }
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->cond, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->true, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->false, offset + 2);
}

PRIVATE inline void PrintArray(FILE *out, UNUSED(Node *node), arrayNode *u, int offset, UNUSED(Bool norecurse))
{
     fputs("Array: ", out);
     if (u->type) {
	  PrintCRSpaces(out, offset + 2);
	  PrintNode(out, u->type, offset + 2);
     }
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->name, offset + 2);
     PrintCRSpaces(out, offset + 2);
     fputs("List: dims", out);
     PrintCRSpaces(out, offset + 4);
     PrintList(out, u->dims, offset + 4);
}

PRIVATE inline void PrintCall(FILE *out, UNUSED(Node *node), callNode *u, int offset, UNUSED(Bool norecurse))
{
     fputs("Call: ", out);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->name, offset + 2);
     PrintCRSpaces(out, offset + 2);
     fputs("List: args", out);
     PrintCRSpaces(out, offset + 4);
     PrintList(out, u->args, offset + 4);
}

PRIVATE inline void PrintInitializer(FILE *out, UNUSED(Node *node), initializerNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Initializer: List: exprs");
     PrintCRSpaces(out, offset + 2);
     PrintList(out, u->exprs, offset + 2);
}

PRIVATE inline void PrintImplicitCast(FILE *out, UNUSED(Node *node), implicitcastNode *u, int offset, UNUSED(Bool norecurse))
{
     fputs("ImplicitCast: ", out);
     if (u->type) {
	  PrintCRSpaces(out, offset + 2);
	  fputs("Type:", out);
	  PrintCRSpaces(out, offset + 4);
	  PrintNode(out, u->type, offset + 2);
     }
     if (u->value) {
	  PrintCRSpaces(out, offset + 2);
	  fputs("Value: ", out);
	  PrintCRSpaces(out, offset + 4);
	  PrintNode(out, u->value, offset + 4);
     }
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->expr, offset + 2);
}

/*************************************************************************/
/*                                                                       */
/*                          Statement nodes                              */
/*                                                                       */
/*************************************************************************/

PRIVATE inline void PrintLabel(FILE *out, Node *node, labelNode *u, UNUSED(int offset), UNUSED(Bool norecurse))
{
     fprintf(out, "Label: %s (%d)", u->name, node->print_uid);
#if 0
     fprintf(out, "Label: %s", u->name);
#endif
}

PRIVATE inline void PrintSwitch(FILE *out, Node *node, SwitchNode *u, int offset, UNUSED(Bool norecurse))
{
     ListMarker marker;
     Node *cse;

     fprintf(out, "Switch: (%d)", node->print_uid);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->expr, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->stmt, offset + 2);
     PrintCRSpaces(out, offset + 2);

     IterateList(&marker, u->cases);
     fprintf(out, "Cases:");
     while (NextOnList(&marker, (GenericREF) & cse)) {
	  fprintf(out, " %d", cse->coord.line);
     }
}

PRIVATE inline void PrintCase(FILE *out, UNUSED(Node *node), CaseNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Case: (container = %d)", u->container->print_uid);
#if 0
     fprintf(out, "Case: ");
#endif
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->expr, offset + 2);
     PrintCRSpaces(out, offset + 2);
}

PRIVATE inline void PrintDefault(FILE *out, UNUSED(Node *node), DefaultNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Default: (container = %d)", u->container->print_uid);
     PrintCRSpaces(out, offset + 2);
}

PRIVATE inline void PrintIf(FILE *out, UNUSED(Node *node), IfNode *u, int offset, UNUSED(Bool norecurse))
{
     fputs("If: ", out);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->expr, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->stmt, offset + 2);
}

PRIVATE inline void PrintIfElse(FILE *out, UNUSED(Node *node), IfElseNode *u, int offset, UNUSED(Bool norecurse))
{
     fputs("IfElse: ", out);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->expr, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->true, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->false, offset + 2);
}

PRIVATE inline void PrintWhile(FILE *out, Node *node, WhileNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "While: (%d) ", node->print_uid);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->expr, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->stmt, offset + 2);
}

PRIVATE inline void PrintDo(FILE *out, Node *node, DoNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Do: (%d) ", node->print_uid);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->stmt, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->expr, offset + 2);
}

PRIVATE inline void PrintFor(FILE *out, Node *node, ForNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "For: (%d) ", node->print_uid);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->init, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->cond, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->next, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->stmt, offset + 2);
}

PRIVATE inline void PrintGoto(FILE *out, UNUSED(Node *node), GotoNode *u, UNUSED(int offset), UNUSED(Bool norecurse))
{
     fprintf(out, "Goto: %s",
	     (u->label ? u->label->u.label.name : "nil"));
}

PRIVATE inline void PrintContinue(FILE *out, UNUSED(Node *node), ContinueNode *u, UNUSED(int offset), UNUSED(Bool norecurse))
{
     fprintf(out, "Continue: (container = %d)", u->container->print_uid);
#if 0
     fprintf(out, "Continue: ");
#endif
}

PRIVATE inline void PrintBreak(FILE *out, UNUSED(Node *node), BreakNode *u, UNUSED(int offset), UNUSED(Bool norecurse))
{
     fprintf(out, "Break: (container = %d)", u->container->print_uid);
#if 0
     fprintf(out, "Break: ");
#endif
}

PRIVATE inline void PrintReturn(FILE *out, UNUSED(Node *node), ReturnNode *u, int offset, Bool UNUSED(norecurse))
{
     fputs("Return: ", out);
#if 0
     if (u->expr) {
	  PrintCRSpaces(out, offset + 2);
	  PrintNode(out, NodeDataType(u->expr), offset + 2);
     }
#endif
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->expr, offset + 2);
}

PRIVATE inline void PrintBlock(FILE *out, UNUSED(Node *node), BlockNode *u, int offset, UNUSED(Bool norecurse))
{
     fputs("Block:", out);
     if (node->parenthesized) fprintf(out, " parenthesized");
     PrintCRSpaces(out, offset + 2);
     if (u->type) {
	 fprintf(out, "Type: (%d)", u->type->print_uid);
     } else {
	 fprintf(out, "Type: nil");
     }
#if 0
     fputs("Type:", out);
#endif
     PrintCRSpaces(out, offset + 4);
     PrintNode(out, u->type, offset + 4);
     PrintCRSpaces(out, offset + 2);
     fputs("List: decl", out);
     PrintCRSpaces(out, offset + 4);
     PrintList(out, u->decl, offset + 4);

     PrintCRSpaces(out, offset + 2);
     fputs("List: stmts", out);
     PrintCRSpaces(out, offset + 4);
     PrintList(out, u->stmts, offset + 4);
}


/*************************************************************************/
/*                                                                       */
/*                            Type nodes                                 */
/*                                                                       */
/*************************************************************************/

PRIVATE inline void PrintPrim(FILE *out, Node *node, UNUSED(primNode *u), UNUSED(int offset), UNUSED(Bool norecurse))
{
     fprintf(out, "Prim: ");
     PrintPrimType(out, node, 0);
     PrintCoord(out, &node->coord);
}

PRIVATE inline void PrintTdef(FILE *out, Node *node, tdefNode *u, UNUSED(int offset), UNUSED(Bool norecurse))
{
     /* temp fix for core dump when printing ast  -Angelina */
     if( u->type == NULL ) 
          fprintf(out, "Tdef: %s (%d) (type=NULL)  ", u->name, node->print_uid);
     else 
          fprintf(out, "Tdef: %s (%d) (type=%d)  ", u->name, node->print_uid, u->type->print_uid); 
     PrintCoord(out, &node->coord);
     PrintTQ(out, u->tq);
#if 0
     if (!norecurse) {
	  PrintCRSpaces(out, offset + 2);
	  PrintNode(out, u->type, offset + 2);
     }
#endif
}

PRIVATE inline void PrintPtr(FILE *out, UNUSED(Node *node), ptrNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Ptr: ");
     PrintTQ(out, u->tq);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->type, offset + 2);
}

PRIVATE inline void PrintAdcl(FILE *out, Node *node, adclNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Adcl: (%d)", node->print_uid);
     PrintTQ(out, u->tq);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->type, offset + 2);
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->dimp->dim, offset + 2);
     if (u->dimp->size > 0) {
	  PrintCRSpaces(out, offset + 2);
	  fprintf(out, "%d", u->dimp->size);
     }
}

PRIVATE inline void PrintFdcl(FILE *out, Node *node, fdclNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Fdcl:");
     PrintCoord(out, &node->coord);
     fprintf(out, " ");
     PrintTQ(out, u->tq);
     PrintCRSpaces(out, offset + 2);
     fputs("List: Args: ", out);
     PrintCRSpaces(out, offset + 4);
     PrintList(out, u->args, offset + 4);
     PrintCRSpaces(out, offset + 2);
     fputs("Returns: ", out);
     PrintCRSpaces(out, offset + 4);
     PrintNode(out, u->returns, offset + 4);
}

PRIVATE inline void PrintSdcl(FILE *out, Node *node, sdclNode *u, int offset, Bool norecurse)
{
     fprintf(out, "Sdcl: (%d) ", node->print_uid);

     if (norecurse)
	  fprintf(out, "%s\n", u->type->name);
     else {
	  PrintCRSpaces(out, offset + 2);
	  PrintTQ(out, u->tq);
	  PrintSUE(out, u->type, offset + 4, TRUE);
     }
}

PRIVATE inline void PrintUdcl(FILE *out, Node *node, udclNode *u, int offset, Bool norecurse)
{
     fprintf(out, "Udcl: (%d) ", node->print_uid);

     if (norecurse)
	  fprintf(out, "%s\n", u->type->name);
     else {
	  PrintCRSpaces(out, offset + 2);
	  PrintTQ(out, u->tq);
	  PrintSUE(out, u->type, offset + 4, TRUE);
     }
}

PRIVATE inline void PrintEdcl(FILE *out, Node *node, edclNode *u, int offset, Bool norecurse)
{
     fprintf(out, "Edcl: (%d) ", node->print_uid);

     if (norecurse)
	  fprintf(out, "%s\n", u->type->name);
     else {
	  PrintCRSpaces(out, offset + 2);
	  PrintTQ(out, u->tq);
	  PrintSUE(out, u->type, offset + 4, TRUE);
     }
}

/*************************************************************************/
/*                                                                       */
/*                      Other nodes (decls et al.)                       */
/*                                                                       */
/*************************************************************************/

PRIVATE inline void PrintDecl(FILE *out, Node *node, declNode *u, int offset, Bool norecurse)
{
     fprintf(out, "Decl: %s (%d)", u->name ? u->name : "", node->print_uid);
     PrintCoord(out, &node->coord);
     fprintf(out, " ");
     PrintTQ(out, u->tq);

     if (norecurse)
	  fprintf(out, "\n");
     else {
	  PrintCRSpaces(out, offset + 2);
	  PrintNode(out, u->type, offset + 2);
	  PrintCRSpaces(out, offset + 2);
	  PrintNode(out, u->init, offset + 2);
	  PrintCRSpaces(out, offset + 2);
	  PrintNode(out, u->bitsize, offset + 2);
	  if (u->attribs) {
	      PrintCRSpaces(out, offset+2);
	      PrintList(out, u->attribs, offset+2);
	  }
     }
}

PRIVATE inline void PrintAttrib(FILE *out, UNUSED(Node *node), attribNode *u,
				int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Attrib: %s", u->name);

     PrintCRSpaces(out, offset + 2);
     PrintList(out, u->arglist, offset + 2);
}

PRIVATE inline void PrintProc(FILE *out, UNUSED(Node *node), procNode *u, UNUSED(int offset), UNUSED(Bool norecurse))
{
     fputs("Proc:\n  ", out);
     PrintNode(out, u->decl, 2);
     fputs("\n  ", out);
     PrintNode(out, u->body, 2);
}

PRIVATE inline void PrintText(FILE *out, UNUSED(Node *node), textNode *u, UNUSED(int offset), UNUSED(Bool norecurse))
{
     fputs("Text: ", out);
     if (u->start_new_line)
	  fputs("(new line) ", out);
     PrintString(out, u->text);
}

/*************************************************************************/
/*                                                                       */
/*                            GCC asm Extensions                         */
/*                                                                       */
/*************************************************************************/

PRIVATE inline void PrintAsm(FILE *out, UNUSED(Node *node), asmNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Asm: ");

     if (u->tq) {
	  PrintCRSpaces(out, offset + 2);
	  PrintTQ(out, u->tq);
     }
     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->template, offset + 2);

     PrintCRSpaces(out, offset + 2);
     fputs("List: output operands", out);
     PrintCRSpaces(out, offset + 4);
     PrintList(out, u->output, offset + 4);

     PrintCRSpaces(out, offset + 2);
     fputs("List: input operands", out);
     PrintCRSpaces(out, offset + 4);
     PrintList(out, u->input, offset + 4);

     PrintCRSpaces(out, offset + 2);
     fputs("List: clobbered registers", out);
     PrintCRSpaces(out, offset + 4);
     PrintList(out, u->clobbered, offset + 4);
}
PRIVATE inline void PrintAsmArg(FILE *out, UNUSED(Node *node), asmargNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "AsmArg: ");
     PrintCRSpaces(out, offset + 2);
     fputs("Constraint:", out);
     PrintNode(out, u->constraint, offset + 2);
     PrintCRSpaces(out, offset + 2);
     fputs("Expr:", out);
     PrintNode(out, u->expr, offset + 2);
}

/*************************************************************************/
/*                                                                       */
/*                            GCC stdarg.h support                       */
/*                                                                       */
/*************************************************************************/
PRIVATE inline void PrintBuiltinVaArg(FILE *out, UNUSED(Node *node), builtinvaargNode *u, int offset, UNUSED(Bool norecurse))
{
    fprintf(out, "builtinvaarg: ");
    PrintCRSpaces(out, offset + 2);
    fputs("Expr:", out);
    PrintNode(out, u->expr, offset + 2);
    PrintCRSpaces(out, offset + 2);
    fputs("Type:", out);
    PrintNode(out, u->type, offset + 2);
}

/*************************************************************************/
/*                                                                       */
/*                            Extensions                                 */
/*                                                                       */
/*************************************************************************/

PRIVATE inline void PrintSpawn(FILE *out, UNUSED(Node *node), spawnNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "Spawn: ");

     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->receiver, offset + 2);

     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->name, offset + 2);

     PrintCRSpaces(out, offset + 2);
     fputs("List: args", out);
     PrintCRSpaces(out, offset + 4);
     PrintList(out, u->args, offset + 4);
}

PRIVATE inline void PrintSync(FILE *out, UNUSED(Node *node), UNUSED(syncNode *u), UNUSED(int offset), UNUSED(Bool norecurse))
{
     fprintf(out, "Sync");
}

PRIVATE inline void PrintInletCall(FILE *out, UNUSED(Node *node), inletcallNode *u, int offset, UNUSED(Bool norecurse))
{
     fprintf(out, "InletCall: ");

     PrintCRSpaces(out, offset + 2);
     PrintNode(out, u->name, offset + 2);

     PrintCRSpaces(out, offset + 2);
     fputs("List: args", out);
     PrintCRSpaces(out, offset + 4);
     PrintList(out, u->args, offset + 4);
}

PRIVATE inline void PrintAbort(FILE *out, UNUSED(Node *node), UNUSED(abortNode *u), UNUSED(int offset), UNUSED(Bool norecurse))
{
     fprintf(out, "Abort");
}

PRIVATE inline void PrintSynched(FILE *out, UNUSED(Node *node), UNUSED(synchedNode *u), UNUSED(int offset), UNUSED(Bool norecurse))
{
     fprintf(out, "Synched");
}


PRIVATE inline void PrintXBlock(FILE *out, UNUSED(Node *node), XBlockNode *u, int offset, UNUSED(Bool norecurse))
{
     fputs("XBlock:", out);
     if (node->parenthesized) fprintf(out, " parenthesized");
     PrintCRSpaces(out, offset + 2);
     if (u->type) {
	 fprintf(out, "Type: (%d)", u->type->print_uid);
     } else {
	 fprintf(out, "Type: nil");
     }
#if 0
     fputs("Type:", out);
#endif
     PrintCRSpaces(out, offset + 4);
     PrintNode(out, u->type, offset + 4);
     PrintCRSpaces(out, offset + 2);
     fputs("List: decl", out);
     PrintCRSpaces(out, offset + 4);
     PrintList(out, u->decl, offset + 4);

     PrintCRSpaces(out, offset + 2);
     fputs("List: stmts", out);
     PrintCRSpaces(out, offset + 4);
     PrintList(out, u->stmts, offset + 4);
}


/*************************************************************************/
/*                                                                       */
/*                      PrintNode and PrintList                          */
/*                                                                       */
/*************************************************************************/

GLOBAL short PassCounter = 0;
GLOBAL int PrintInvocations = 0;	/* number of pending PrintNodes on call
					 * stack */

static void kill_print_uid (Node *n) {
    n->print_uid = -1;
    ASTWALK(n, kill_print_uid);
}

static int print_uid = 0;
static void assign_uids (Node *n) {
    if (n->print_uid==-1) {
	n->print_uid = print_uid++;
	ASTWALK(n, assign_uids);
    }
}


GLOBAL void PrintWildcard(FILE *out, Node *node, int offset) {
  
  PrintCRSpaces(out, offset);
  switch (node->wTyp) {
  case WildcardE:  
    fprintf(out, "WildcardE: ");
    break;
  case WildcardD:  
    fprintf(out, "WildcardD: ");
    break;
  case WildcardT:
    fprintf(out, "WildcardT: ");
    break;
  default:
    fprintf(out, "Unknown wildcard.");
  }
}



GLOBAL void PrintNode(FILE *out, Node *node, int offset)
{
  Bool norecurse;

  if (node == NULL) {
    fprintf(out, "nil");
    return;
  }
  if (PrintInvocations++ == 0) {
    /* then we're the first invocation for this pass over the tree */
    ++PassCounter;
    print_uid=MaxBasicType; /* The first MaxBasicType values are used to uniquely identify prims.  Ugly...*/
    kill_print_uid(node);
    assign_uids(node);
  }
  norecurse = (node->pass == PassCounter);
  node->pass = PassCounter;

  if (IsWildcard(node)) {
    PrintWildcard(out, node, offset);
  }
  
#define CODE(name, node, union) Print##name(out,node,union,offset,norecurse)
  ASTSWITCH(node, CODE)
#undef CODE

    if (node->analysis.livevars) {
      PrintCRSpaces(out, offset + 2);
      PrintAnalysis(out, node);
    }
  --PrintInvocations;
}

GLOBAL void PrintList(FILE *out, List *list, int offset)
{
     ListMarker marker;
     Node *item;
     Bool firstp = TRUE;

     if (PrintInvocations++ == 0) {
	  /* then we're the first invocation for this pass over the tree */
	  ++PassCounter;
	  print_uid=MaxBasicType;
	  LISTWALK(list, kill_print_uid);
	  LISTWALK(list, assign_uids);
     }
     IterateList(&marker, list);
     while (NextOnList(&marker, (GenericREF) & item)) {
	  if (firstp == TRUE)
	       firstp = FALSE;
	  else if (offset < 0)
	       fputs("\n\n", out);
	  else
	       PrintCRSpaces(out, offset);

	  PrintNode(out, item, offset);
     }

     if (firstp == TRUE)
	  fputs("nil", out);

     --PrintInvocations;
}

/*************************************************************************/
/*                                                                       */
/*                      Low-level output routines                        */
/*                                                                       */
/*************************************************************************/

int print_float(FILE *fd, float val)
{
     int i;
     char fmt[8];
     char buf[64];
     float tmp;

     i = 7;
     while (1) {
	  sprintf(fmt, "%%.%dg", i);
	  sprintf(buf, fmt, val);
	  sscanf(buf, "%f", &tmp);
	  if (tmp == val)
	       break;
	  i += 1;
	  assert(i < 20);
     }

     return fprintf(fd, "%s", buf);
}

int print_double(FILE *fd, double val)
{
     int i;
     char fmt[8];
     char buf[64];
     double tmp;

     i = 16;
     while (1) {
	  sprintf(fmt, "%%.%dlg", i);
	  sprintf(buf, fmt, val);
	  sscanf(buf, "%lf", &tmp);
	  if (tmp == val)
	       break;
	  i += 1;
	  assert(i < 20);
     }

     return fprintf(fd, "%s", buf);
}

GLOBAL int PrintConstant(FILE *out, Node *c, Bool with_name)
{
     int len = 0;

     if (with_name)
	  switch (c->u.Const.type->typ) {
	      case Prim:
		   len = PrintPrimType(out, c->u.Const.type, 0) + 1;
		   fputc(' ', out);
		   break;
		   /* Used for strings */
	      case Adcl:
		   assert(c->u.Const.type->u.adcl.type->typ == Prim);
		   fprintf(out, "array of ");
		   len = PrintPrimType(out, c->u.Const.type->u.adcl.type, 0) + 10;
		   fputc(' ', out);
		   break;
	      default:
		   len = fprintf(out, "??? ");
	  }
     switch (c->u.Const.type->typ) {
	 case Prim:
	      switch (c->u.Const.type->u.prim.basic) {
		  case Sint:
		       return len + fprintf(out, "%d", c->u.Const.value.i);

		       /*Manish 2/3 hack to print pointer constants */
		  case Uint:
		       return len + fprintf(out, "%uU", c->u.Const.value.u);
		  case Slong:
		       return len + fprintf(out, "%ldL", c->u.Const.value.l);
		  case Ulong:
		       return len + fprintf(out, "%luUL", c->u.Const.value.ul);
		  case Float:
		       return len + print_float(out, c->u.Const.value.f);
		  case Double:
		       return len + print_double(out, c->u.Const.value.d);
		  case Char:
		  case Schar:
		  case Uchar:
		       return len + PrintChar(out, c->u.Const.value.i);

		  default:
		       Fail(__FILE__, __LINE__, "");
		       return 0;
	      }

	      /* Manish 2/3  Print Constant Pointers */
	 case Ptr:
	      return len + fprintf(out, "%u", c->u.Const.value.u);
	      /* Used for strings */
	 case Adcl:
	      return len + PrintString(out, c->u.Const.value.s);

	 default:
	     FAIL("Unrecognized constant type");
     }
}

void PrintCRSpaces(FILE *out, int spaces)
{
     fputc('\n', out);
     while (spaces--)
	  fputc(' ', out);
}

void PrintSpaces(FILE *out, int spaces)
{
     while (spaces--)
	  fputc(' ', out);
}

GLOBAL void CharToText(char *array, unsigned char value)
{
     if (value < ' ') {
	  static const char *names[32] =
	  {
	       "nul", "soh", "stx", "etx", "eot", "enq", "ack", "bel",
	       "\\b", "\\t", "\\n", "\\v", "ff", "cr", "so", "si",
	       "dle", "dc1", "dc2", "dc3", "dc4", "nak", "syn", "etb",
	       "can", "em", "sub", "esc", "fs", "gs", "rs", "us"};
	  sprintf(array, "0x%02x (%s)", value, names[value]);
     } else if (value < 0x7f) {
	  sprintf(array, "'%c'", value);
     } else if (value == 0x7f) {
	  strcpy(array, "0x7f (del)");
     } else {			/* value >= 0x80 */
	  sprintf(array, "0x%x", value);
     }
}

GLOBAL inline int PrintChar(FILE *out, int value)
{
     switch (value) {
	 case '\n':
	      return fprintf(out, "\\n");
	 case '\t':
	      return fprintf(out, "\\t");
	 case '\v':
	      return fprintf(out, "\\v");
	 case '\b':
	      return fprintf(out, "\\b");
	 case '\r':
	      return fprintf(out, "\\r");
	 case '\f':
	      return fprintf(out, "\\f");
	 case '\a':
	      return fprintf(out, "\\a");
	 case '\\':
	      return fprintf(out, "\\\\");
	 case '\?':
	      return fprintf(out, "\\\?");
	 case '\"':
	      return fprintf(out, "\\\"");
	 case '\'':
	      return fprintf(out, "\\\'");
	 default:
	      if (isprint(value)) {
		   fputc(value, out);
		   return 1;
	      } else {
		   return fprintf(out, "\\%o", value);
	      }
     }
}

GLOBAL int PrintString(FILE *out, const char *s)
{
     int len = 0;

     fputc('\"', out);
     while (*s != 0) {
	  len += PrintChar(out, *s++);
     }
     fputc('\"', out);

     return len + 2;
}
