/* This program models the microbenchmark used in \cite{RawjarGo01}
 *  (Rajwar and Goodman, "Speculative Lock Elision...", MICRO-34.)
 *
 * In that microbenchmark there are N threads, each of which does the
 * following increments a thread-local counter (2^16)/N times, and all
 * N counters are prtected by a single global lock.
 *
 *  do_thread(int threadnum) {
 *    for (i=0; i<(1<<16)/N; i++) {
 *       lock(global_lock);
 *       thread_local[threadnum]++;
 *       unlock(global_lock);
 *    }
 *  }
 */
#include <stdlib.h>
#include <assert.h>

extern void __c2c_ProfileRead(void*, int);
extern void __c2c_ProfileWrite(void*, int);

struct counter {
    int the_counter;
    char padding[64-sizeof(int)];
};

int main (int argc, char *argv[]) {
    assert(argc==2);
    {
	int N = atoi(argv[1]);
	struct counter *counters=malloc(N*sizeof(struct counter));
	{
	    int p;
	    for (p=0; p<N; p++) counters[p].the_counter=0;
	}
	{
	    int i;
	    int p=0;
	    for (i=0; i<(1<<16); i++) {
		counters[p++].the_counter++;
		if (p>=N) p=0;
	    }
	}
	free(counters);
    }
    return 0;
}
