/****************************************************************************/
/* UTIL: various useful routines and functions.                             */
/*                                                                          */
/* Copyright (c) 1993 by Joshua E. Barnes, Honolulu, HI.                    */
/* It's free because it's yours.                                            */
/****************************************************************************/
static const char *ident __attribute__((__unused__))
     = "$HeadURL: https://bradley.lcs.mit.edu/svn/repos/cilk/5.4.2.2/examples/Barnes-Hut/util.c $ $LastChangedBy: bradley $ $Rev: 1708 $ $Date: 2003-02-27 13:25:07 -0500 (Thu, 27 Feb 2003) $";


#include "stdinc.h"
#include "real.h"
#include "vectmath.h"
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/param.h>

#include <cilk-lib.h>
#include <cilk-compat.h>

#ifndef HZ
#  include <time.h>
#  define HZ CLK_TCK
#endif

void error(string, ...);

extern double drand48(void);			/* should be in math.h      */

/*
 * XRANDOM: generate floating-point random number.
 */

real xrandom(real xl, real xh)
{
    return (xl + (xh - xl) * drand48());
}

/*
 * RSQR: compute x*x.
 */

real rsqr(real x)
{
    return (x * x);
}

/*
 * DISTV: subtract vectors and return distance between.
 */

real distv(vector v, vector u)
{
    real s, d;
    int n = NDIM;

    s = 0.0;
    while (--n >= 0) {
	d = (*v++) - (*u++);
	s += d * d;
    }
    return (rsqrt(s));
}

/*
 * STREQ: test for equality of strings.
 */

bool streq(string a, string b)
{
    return (strcmp(a, b) == 0);
}

/*
 * SCANOPT: scan string of the form "word1,word2,..." for match. Warning:
 * words must be separated by exactly one comma -- no spaces allowed!
 */

bool scanopt(string opt, string key)
{
    char *op, *kp;

    op = (char *) opt;				/* start scan of options    */
    while (*op != '\0') {			/* loop over words in opt   */
	kp = (char *) key;			/*   start at front of key  */
	while ((*op != ',' ? *op : '\0')	/*   loop while this word   */
	         == *kp) {			/*   ...matches text of key */
	    if (*kp++ == '\0')			/*     at end of key word?  */
		return (TRUE);			/*       keyword found      */
	    op++;				/*     else advance ptrs    */
	}
	while (*op != '\0' && *op++ != ',')	/*   loop till end of word, */
	    continue;				/*     passing "," at end   */
    }
    return (FALSE);				/* keyword not found        */
}

/*
 * CPUTIME: compute CPU time in minutes.
 */

real cputime()
{
    struct tms buffer;

    if (times(&buffer) == -1)
	error("times() call failed\n");
    return (buffer.tms_utime / (60.0 * HZ));
}

/*
 * ALLOCATE: memory allocation with error checking.
 */

void *allocate(nb)
int nb;
{
    void *mem;

    mem = (void *) calloc(nb, 1);		/* calloc zeros memory      */
    if (mem == NULL)
	error("allocate: not enuf memory (%d bytes)\n", nb);
    return (mem);
}

/*
 * ERROR: print error message and exit.
 */

void error(string msg, ...)
{
    va_list args;

    va_start(args, msg);
    vfprintf(stderr, msg, args);
    va_end(args);
    exit(-1);					/* quit with error status   */
}

/*
 * EPRINTF: print error message, but don't exit.
 */

void eprintf(string msg, ...)
{
    va_list args;

    va_start(args, msg);
    vfprintf(stderr, msg, args);
    va_end(args);
}
