#!/usr/local/bin/perl -w

# output directories
##########################
$abi_dir_n32 = "mips-abi-n32";
$abi_dir_o32 = "mips-abi-o32";
$abi_dir_64 = "mips-abi-64";

$mipslib_dir="mipslib";


# file lists for the 2 libraries to be compiled for run-time 
# linkage with Cil programs
############################################################
@libcilkrt0_files=("cilkrt0.c");
@libcilk_files=("barrier.c","debug.c","timing.c","cilk-main.c","hooks.c","mutex.c","workers.c","cilk.c","internal-malloc.c","sched.c","invoke-main.c","stats.c","cmdline.c","malloc.c","sysdep.c", "gcc-builtin.c");

#preprocessor flags for debug/profiling versions
################################################
$debug_macro = "-DCILK_DEBUG=1";
$profile_macro = "-DCILK_TIMING=1 -DCILK_STATS=1 -DCILK_CRITICAL_PATH=1";




# a subroutine to compile a list of files. it returns the list of 
# object files.
##########################################
sub compile_c_file {
    # define local variables & extract argument from arg list
    #########################################################
    my($out_dir, $abi_flag, $debug_flag, $extra_flags, @files_list);
    ($out_dir, $abi_flag, $debug_flag, $extra_flags, @files_list)=@_;
    my($out_file, @out_file_list);

    foreach $file_name (@files_list) {
	# make 'file.c' into 'file.o'
	$out_file=$file_name;
	chop($out_file);
	$out_file="${out_file}o";
	# add the output file into the returned list
	@out_file_list=(@out_file_list , "${out_dir}\/${out_file}");

	# prepare the mipspro command line
	$cmd_line="cc -I. -I.. -DHAVE_CONFIG_H -DSGI_MIPSPRO ${abi_flag} ${debug_flag} -o ${out_dir}\/${out_file} -woff 1155 ${extra_flags} -c ${file_name}";
	print("${cmd_line}\n");
	# execute it 
	system($cmd_line);
    }
    return @out_file_list;
}


# a subroutine to link a list of files & generate a library 
# archive
##########################################
sub archive_files {
    # define local variables & extract argument from arg list
    #########################################################
    my($out_dir, $out_file, @obj_list,$cmd_line);
    ($out_dir, $out_file, @obj_list)=@_;
    #print("inputs: $out_dir <> $out_file <> @obj_list\n");

    # prepare the mipspro command line
    $cmd_line="cc -ar -o ${out_dir}\/${out_file} @{obj_list}";
    print("${cmd_line}\n");
    # execute it 
    system($cmd_line);
}

# delete all object files in a directory.
#########################################
sub delete_all_obj_files {
    # define local variables & extract argument from arg list
    #########################################################
    my($dir, @obj_files);
    ($dir)=@_;

    # delete object files from output directory.
    @obj_files = "${dir}\/*.o";
    print("removing @{obj_files}\n");
    unlink <@obj_files>;
}

# delete all archives
######################
sub delete_all_archive_files {
    # define local variables & extract argument from arg list
    #########################################################
    my($dir, @ar_files);
    ($dir)=@_;

    # delete archives files from directory.
    @ar_files = "${dir}\/*.a";
    print("removing @{ar_files}\n");
    unlink <@ar_files>;
}

# create a directory if doesn't exist
#######################################
sub make_dir {
    # define local variables & extract argument from arg list
    #########################################################
    my($dir);
    ($dir) = @_;
    if (-d $dir) {
	print("directory ${dir} already exists\n");
    } else {
	print("creating directory ${dir}\n");
	mkdir($dir, 0700);
    }
}

# make 4 versions of 2 libraries for the ABI
############################################
sub build_lib {
    # define local variables & extract arguments from arg list
    #########################################################
    my($out_dir, $out_lib, $abi_flag, $debug_flag, 
       $profile_glag, $extra_flags, @files_list);
    ($out_dir, $out_lib, $abi_flag, $debug_flag, 
     $profile_glag, $extra_flags, @files_list)=@_;

    delete_all_obj_files($out_dir);
    print("compiling ${abi_flag} ABI ${out_lib} ...\n");
    print("debug is ${debug_flag}\n");
    print("profile is ${profile_glag}\n");
    if (length($debug_flag) != 0) {
	# debug_flag used
	$extra_flags = "${extra_flags} ${debug_macro}";
	printf("fixing extra_flags ${extra_flags}\n");
    }
    if ($profile_glag == 1) {
	# profiling is used
	$extra_flags = "${extra_flags} ${profile_macro}";
	print("fixing extra_flags ${extra_flags}\n");
    }
    @out_file_list = compile_c_file($out_dir, 
				    $abi_flag,
				    $debug_flag, 
				    $extra_flags, 
				    @files_list);
#print("out_file_list: @out_file_list\n");
    archive_files($out_dir, $out_lib, @out_file_list);

    # put a link in the "runtime/mipslibs" directory
    #$lib_name = "${out_dir}\/${out_lib}";
    #$link_file = "${mipslib_dir}\/${out_lib}";
    #print("lib_name = $lib_name , link_file = $link_file\n");
    #symlink ($lib_name, $link_file);
}

# build a complete ABI, and set create a link "runtime/mipslib"
# to that runtime/<abi-specific-dir>
################################################################
sub build_abi {
    # define local variables & extract arguments from arg list
    #########################################################
    my($abi_dir, $abi_flag);
    ($abi_dir, $abi_flag)=@_;

    make_dir($abi_dir);
    delete_all_archive_files($abi_dir);
    if (-d $mipslib_dir) {
	print("removing ${mipslib_dir}\n");
	unlink <$mipslib_dir>;
    }
    
    
    # debug = no , profiling = no
    build_lib($abi_dir, "libcilkrt0.a", $abi_flag, "", 
	      0, "-Ofast", @libcilkrt0_files);
    build_lib($abi_dir, "libcilk.a", $abi_flag, "", 
	      0, "-woff 3303 -Ofast", @libcilk_files);

    #debug = yes , profiling = no
    build_lib($abi_dir, "libcilkrt0.g.a", $abi_flag, "-g3", 
	      0, "", @libcilkrt0_files);
    build_lib($abi_dir, "libcilk.g.a", $abi_flag, "-g3", 
	      0, "-woff 3303", @libcilk_files);
    
    #debug = no , profiling = yes
    build_lib($abi_dir, "libcilkrt0.p.a", $abi_flag, "",
	      1, "-Ofast", @libcilkrt0_files);
    build_lib($abi_dir, "libcilk.p.a", $abi_flag, "", 
	      1, "-woff 3303 -Ofast", @libcilk_files);

    #debug = yes , profiling = yes
    build_lib($abi_dir, "libcilkrt0.g.p.a", $abi_flag, "-g3",
	      1, "-woff 3303", @libcilkrt0_files);
    build_lib($abi_dir, "libcilk.g.p.a", $abi_flag, "-g3",
	      1, "-woff 3303", @libcilk_files);

    # link in the "runtime/mipslibs" directory to the current ABI dir
    print("linking ${abi_dir} to ${mipslib_dir}\n");
    symlink ($abi_dir, $mipslib_dir);
}


# n32 ABI
###########
build_abi ($abi_dir_n32, "-n32");
print("Not compiling o32 & 64 bit ABI's\n");
die();

# o32 ABI
###########
build_abi ($abi_dir_o32, "-o32");


# 64 ABI
##########
build_abi ($abi_dir_64, "-64");


