/* Time fbmfence vs.xchgl for a memory barrier. */
#include <sys/time.h>
#include <stdio.h>

enum { COUNT = 100000000 };

   static inline void CILK_MB(void)
   {
	/*
	 * We use an xchg instruction to serialize memory accesses, as can
	 * be done according to the Intel Architecture Software Developer's
	 * Manual, Volume 3: System Programming Guide
	 * (http://www.intel.com/design/pro/manuals/243192.htm), page 7-6,
	 * "For the P6 family processors, locked operations serialize all
	 * outstanding load and store operations (that is, wait for them to
	 * complete)."  The xchg instruction is a locked operation by
	 * default.  Note that the recommended memory barrier is the cpuid
	 * instruction, which is really slow (~70 cycles).  In contrast,
	 * xchg is only about 23 cycles (plus a few per write buffer
	 * entry?).  Still slow, but the best I can find.  -KHR 
	 */
	int x=0, y;
	asm volatile ("xchgl %0,%1" :"=r" (x) :"m" (y), "0" (x) :"memory");
   }

double tdiff (struct timeval *start, struct timeval *end) {
    return ((end->tv_sec-start->tv_sec + 1e-6*(end->tv_usec + start->tv_usec))/COUNT)*1e9;
}

void do1 (int *x) {
    int i;
    struct timeval start, end;
    gettimeofday(&start, 0);
    for (i=0; i<COUNT; i++) {
	x[0]++;
	x[1]++;
	x[2]++;
	x[3]++;
    }
    gettimeofday(&end, 0);
    printf("do1    %6.1fns/loop\n", tdiff(&start, &end));
}

void domfence (int *x) {
    int i;
    struct timeval start, end;
    gettimeofday(&start, 0);
    for (i=0; i<COUNT; i++) {
	x[0]++;
	x[1]++;
	asm volatile ("mfence":::"memory");
	x[2]++;
	x[3]++;
    }
    gettimeofday(&end, 0);
    printf("mfence:%6.1fns/loop\n", tdiff(&start, &end));
}

void doxchgl (int *x) {
    int i;
    struct timeval start, end;
    gettimeofday(&start, 0);
    for (i=0; i<COUNT; i++) {
	x[0]++;
	x[1]++;
	CILK_MB();
	x[2]++;
	x[3]++;
    }
    gettimeofday(&end, 0);
    printf("xchgl: %6.1fns/loop\n", tdiff(&start, &end));
}

int main (UNUSED(int argc),
	  UNUSED(char *argv[]) {
    int x[4];
    do1(x);
    domfence(x);
    doxchgl(x);
    return 0;
}
